"""Note: We used to have a version of this that was quite flexible and could allow for others to provide their capabilities and the Python engine could use them.
However, we have decided to simplify the Python engine to only use the LLM and Embedding from the context for the moment as the Python Engine we want is a bit complex.
We will revisit this in the future to be able to have both:
- Capabilities that can be provided by others
- State management that can allow a complex agentic workflow"""

from lavague.core.context import Context, get_default_context
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.embeddings import BaseEmbedding

import trafilatura
from llama_index.core import Document, VectorStoreIndex

class PythonEngine:

    llm: BaseLLM
    embedding: BaseEmbedding
        
    def __init__(
        self,
        llm: BaseLLM = get_default_context().llm,
        embedding: BaseEmbedding = get_default_context().embedding,
        ):
        self.llm = llm
        self.embedding = embedding

    @classmethod
    def from_context(
        cls,
        context: Context,
    ):
        return cls(context.llm, context.embedding)
        
    def extract_information(self, instruction: str, html: str) -> str:
        llm = self.llm
        embedding = self.embedding
        
        page_content = trafilatura.extract(html)
        # Next we will use Llama Index to perform RAG on the extracted text content

        documents = [Document(text=page_content)]

        # We then build index
        index = VectorStoreIndex.from_documents(documents, embed_model=embedding)
        query_engine = index.as_query_engine(llm=llm)

        # We finally store the output in the variable 'output'
        output = query_engine.query(instruction).response
        
        return output


import os
from typing import Any, Dict, Optional
import requests
import uuid
import msgpack
from PIL import Image
from io import BytesIO

from .version_checker import get_installed_version

TELEMETRY_VAR = os.getenv("LAVAGUE_TELEMETRY")
UNIQUE_ID = os.getenv("LAVAGUE_UNIQUE_USER_ID")
USER_ID = str(uuid.uuid4())

if UNIQUE_ID is not None:
    UNIQUE_ID = UNIQUE_ID[:256]


def compress_img(img: Image):
    buffer: BytesIO = BytesIO()
    img_ret = img.resize((1024, 1024), Image.LANCZOS)
    img_ret = img_ret.convert("RGB")
    img_ret.save(buffer, "PNG", quality=50)
    return buffer.getvalue()


def send_telemetry_scr(
    action_id: str, before: Image, image: Image, after: Image, test: bool = False
):
    try:
        if TELEMETRY_VAR == "HIGH":
            if before is not None:
                before = compress_img(before)
            if image is not None:
                image = compress_img(image)
            if after is not None:
                after = compress_img(after)
            dict_img = {
                "action_id": action_id,
                "before": before,
                "image": image,
                "after": after,
            }
            pack = msgpack.packb(dict_img)
            r = requests.post(
                "https://telemetrylavague.mithrilsecurity.io/telemetry_scrs", data=pack
            )
            if r.status_code != 200:
                raise ValueError(r.content)
        else:
            pass
    except Exception as e:
        if not test:
            print("Telemetry (screenshot) failed with ", e)
        else:
            raise ValueError("Telemetry failed with ", e)


def send_telemetry(
    model_name: str,
    code: str,
    instruction: str,
    url: str,
    origin: str,
    success: bool,
    test: bool = False,
    error: str = "",
    source_nodes: str = "",
    bounding_box: Optional[Dict[str, int]] = None,
    viewport_size: Optional[Dict[str, int]] = None,
    main_objective: str = "",
    objectives: str = "",
    action_id: str = "",
    multi_modal_model: str = "",
    step_id: str = "",
    run_id: str = "",
):
    """
    Telemetry to help performance.
    Mandatory telemetry variables - DO NOT DELETE ANY, else telemetry will fail: model_name, code, screenshot, html, source_nodes, instruction, url, origin, success
    """
    success_str = str(success)
    try:
        if TELEMETRY_VAR is None:
            json_send = {
                "action_id": action_id,
                "version": get_installed_version("lavague"),
                "code_produced": code,
                "llm": model_name,
                "unique_id": UNIQUE_ID,
                "user_id": USER_ID,
                "origin": origin,
                "url": url,
                "success": success_str,
                "instruction": instruction,
                "source_nodes": source_nodes,
                "error_msg": error,
                "test": test,
                "bounding_box": bounding_box,
                "viewport_size": viewport_size,
                "main_objective": main_objective,
                "objectives": objectives,
                "multi_modal_model": multi_modal_model,
                "run_id": run_id,
                "step_id": step_id,
            }
            r = requests.post(
                "https://telemetrylavague.mithrilsecurity.io/telemetry",
                json=json_send,
            )
            if r.status_code != 200:
                raise ValueError(r.content)
        elif TELEMETRY_VAR == "NONE":
            pass
    except Exception as e:
        if not test:
            print("Telemetry failed with ", e)
        else:
            raise ValueError("Telemetry failed with ", e)



from __future__ import annotations
from typing import Optional, Generator, List
from llama_index.core.query_engine import RetrieverQueryEngine
from llama_index.core import get_response_synthesizer, QueryBundle, PromptTemplate
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding
from lavague.core.extractors import BaseExtractor, PythonFromMarkdownExtractor
from lavague.core.retrievers import BaseHtmlRetriever, OpsmSplitRetriever
from lavague.core.base_driver import BaseDriver
from lavague.core.action_template import ActionTemplate
from lavague.core.context import Context, get_default_context

ACTION_ENGINE_PROMPT_TEMPLATE = ActionTemplate(
    """
{driver_capability}

HTML:
{context_str}
Query: {query_str}
Completion:

""",
    PythonFromMarkdownExtractor(),
)


class ActionEngine:
    """
    ActionEngine leverages the llm model and the embedding model to output code from the prompt and the html page.

    Args:
        driver (`BaseDriver`):
            The Web driver used to interact with the headless browser
        llm (`BaseLLM`)
            llama-index LLM that will generate the action
        embedding (`BaseEmbedding`)
            llama-index Embedding model
        retriever (`BaseHtmlRetriever`)
            Specify which algorithm will be used for RAG
        prompt_template (`PromptTemplate`)
            Squelette of the final prompt
        extractor (`BaseExtractor`)
            Specify how to extract the final code from the llm answer
    """

    def __init__(
        self,
        driver: BaseDriver,
        llm: BaseLLM = get_default_context().llm,
        embedding: BaseEmbedding = get_default_context().embedding,
        retriever: BaseHtmlRetriever = OpsmSplitRetriever(),
        prompt_template: PromptTemplate = ACTION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = ACTION_ENGINE_PROMPT_TEMPLATE.extractor,
    ):
        self.driver: BaseDriver = driver
        self.llm: BaseLLM = llm
        self.embedding: BaseEmbedding = embedding
        self.retriever: BaseHtmlRetriever = retriever
        self.prompt_template: PromptTemplate = prompt_template.partial_format(
            driver_capability=driver.get_capability()
        )
        self.extractor: BaseExtractor = extractor

    @classmethod
    def from_context(
        cls,
        context: Context,
        driver: BaseDriver,
        retriever: BaseHtmlRetriever = OpsmSplitRetriever(),
        prompt_template: PromptTemplate = ACTION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = ACTION_ENGINE_PROMPT_TEMPLATE.extractor,
    ) -> ActionEngine:
        """
        Create an ActionEngine from a context
        """
        return cls(
            driver,
            context.llm,
            context.embedding,
            retriever,
            prompt_template,
            extractor,
        )

    def _get_query_engine(self, streaming: bool = True) -> RetrieverQueryEngine:
        """
        Get the llama-index query engine

        Args:
            html: (`str`)
            streaming (`bool`)

        Return:
            `RetrieverQueryEngine`
        """

        response_synthesizer = get_response_synthesizer(
            streaming=streaming, llm=self.llm
        )

        # assemble query engine
        query_engine = RetrieverQueryEngine(
            retriever=self.retriever.to_llama_index(self.driver, self.embedding),
            response_synthesizer=response_synthesizer,
        )

        query_engine.update_prompts(
            {"response_synthesizer:text_qa_template": self.prompt_template}
        )

        return query_engine

    def get_nodes(self, query: str) -> List[str]:
        """
        Get the nodes from the html page

        Args:
            html (`str`): The html page

        Return:
            `List[str]`: The nodes
        """
        source_nodes = self.retriever.retrieve_html(
            self.driver, self.embedding, QueryBundle(query_str=query)
        )
        source_nodes = [node.text for node in source_nodes]
        return source_nodes

    def get_action_from_context(self, context: str, query: str) -> str:
        """
        Generate the code from a query and a context
        """
        prompt = self.prompt_template.format(context_str=context, query_str=query)
        response = self.llm.complete(prompt).text
        code = self.extractor.extract(response)
        return code

    def get_action(self, query: str) -> Optional[str]:
        """
        Generate the code from a query

        Args:
            query (`str`): Instructions given at the end of the prompt to tell the model what to do on the html page

        Return:
            `str`: The generated code
        """
        query_engine = self._get_query_engine(streaming=False)
        response = query_engine.query(query)
        code = response.response
        return self.extractor.extract(code)

    def get_action_streaming(self, query: str) -> Generator[str, None, None]:
        """
        Stream code from a query and an url

        Args:
            query (`str`): Instructions given at the end of the prompt to tell the model what to do on the html page
            url (`str`): The url of the target html page

        Return:
            `Generator[str, None, None]`: Generator for the generated code
        """
        query_engine = self._get_query_engine(streaming=True)
        streaming_response = query_engine.query(query)
        for text in streaming_response.response_gen:
            yield text



import time
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By


class NavigationControl:
    def __init__(self, driver: WebDriver) -> None:
        self.driver = driver

    def scroll_one_viewport(self, direction: Keys):
        viewport_height = self.driver.execute_script("return window.innerHeight")

        body = self.driver.find_element(By.TAG_NAME, "body")
        num_scrolls = (
            viewport_height // 50
        )  # Assuming each arrow key press scrolls 50 pixels
        for _ in range(num_scrolls):
            body.send_keys(direction)
            time.sleep(0.05)

    def execute_instruction(self, instruction):
        if "SCROLL_DOWN" in instruction:
            self.scroll_one_viewport(Keys.ARROW_DOWN)
        elif "SCROLL_UP" in instruction:
            self.scroll_one_viewport(Keys.ARROW_UP)
        elif "WAIT" in instruction:
            time.sleep(2)
        else:
            raise ValueError(f"Unknown instruction: {instruction}")



from __future__ import annotations
from typing import Optional
from abc import ABC
import os

from llama_index.core import PromptTemplate
from llama_index.core.multi_modal_llms import MultiModalLLM

from lavague.core.context import Context, get_default_context

WORLD_MODEL_GENERAL_EXAMPLES = """
Objective:  Go to the first issue you can find
Previous instructions: 
- Click on 'Issues' with the number '28' next to it.
- [FAILED] Click on 'Build and share place where people can suggest their use cases and results #225'
- [FAILED] Click on 'Build and share place where people can suggest their use cases and results #225'
Last engine: Navigation Engine
Current state: 
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: []
  user_inputs: []

Thoughts:
- The current screenshot shows the issues page of the GitHub repository 'lavague-ai/LaVague'.
- The objective is to go to the first issue.
- Previous instructions have been unsuccessful. A new approach should be used.
- The '#225' seems not to be clickable and it might be relevant to devise an instruction that does not include it. 
Next engine: Navigation Engine
Instruction: Click on the first issue, with title 'Build and share place where people can suggest their use cases and results'
-----
Objective: Find When Llama 3 was released
Previous instructions:
- Click on 'meta-llama/Meta-Llama-3-8B'
Last engine: Navigation Engine
Current state:
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: []
  user_inputs: []

Thoughts:
- The current screenshot shows the model page for 'meta-llama/Meta-Llama-3-8B' on Hugging Face.
- Hugging Face, is a hub for AI models and datasets, where users can explore and interact with a variety of AI models.
- I am therefore on the right page to find information about the release date of 'Meta-Llama-3-8B'.
- To find the release date, I need to locate the relevant information in the content of the page.
- Therefore the best next step is to use the Python Engine to extract the release date from the content of the page.
Next engine: Python Engine
Instruction: Extract the release date of 'Meta-Llama-3-8B' from the textual content of the page.
-----
Objective: Provide the code to get started with Gemini API
Previous instructions:
- Click on 'Read API docs'
- Click on 'Gemini API quickstart' on the menu
Last engine: Navigation Engine
Current state:
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: []
  user_inputs: []

Thoughts:
- The current screenshot shows the documentation page for the getting started of Gemini API.
- I am therefore on the right page to find the code to get started with the Gemini API.
- The next step is to provide the code to get started with the Gemini API.
- Therefore I need to use the Python Engine to generate the code to extract the code to get started with the Gemini API from this page.
Next engine: Python Engine
Instruction: Extract the code to get started with the Gemini API from the content of the page.
-----
Objective: Show what is the cheapest product
Previous instructions: [NONE]
Last engine: [NONE]
Current state:
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: []
  user_inputs: []
  
Thoughts:
- The screenshot shows an e-commerce website with various products.
- To find the cheapest product, I need to identify the product with the lowest price.
- There seems to be selectors for sorting products by price on the left side of the page.
- The screenshot only shows part of the selectors for price. I should scroll down to see the full list of products and prices.
Next engine: Navigation Controls
Instruction: SCROLL_DOWN
-----
Objective: What tech stack do we use?
Previous instructions: 
- [FAILED] Locate and click on the "Technology Solutions" link or section to find information about the tech stack.
- [FAILED] Click on the "Technology Solutions" section to explore detailed information about the tech stack.
Last engine: Navigation Engine
Current state: 
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: []
  user_inputs: []

Thought:
- The screenshot shows a Notion webpage with information about a company called ACME INC.
- It has information about the company, their services, and departments.
- Previous instructions tried to click on "Technology Solutions" without success. This probably means that "Technology Solutions" is not clickable or reachable.
- Other strategies have to be pursued to reach the objective.
- There seems to be information at the end of the screen about departments, with mention of a 'Software development' section that could be promising.
- The best next step is to scroll down to gather more information.
Next engine: Navigation Controls
Instruction: SCROLL_DOWN
-----
Objective: Find the description of the author
Previous instructions:
- Click on 'About the author'
- Extract the description of the author from the content of the page.
Last engine: Python Engine
Current state:
external_observations:
  vision: '[SCREEENSHOT]'
internal_state:
  agent_outputs: [
    'The author is a software engineer with a passion for AI and machine learning. He has worked on various projects and has a blog where he shares his knowledge and experience.'
  ]
  user_inputs: []

Thoughts:
- The screenshot shows the description of the author.
- The description of the author has been successfully extracted from the content of the page.
- The objective has been reached.
Next engine: STOP
Instruction: STOP
"""

WORLD_MODEL_PROMPT_TEMPLATE = PromptTemplate(
    """
You are an AI system specialized in high level reasoning. Your goal is to generate instructions for other specialized AIs to perform web actions to reach objectives given by humans.
Your inputs are:
- objective ('str'): a high level description of the goal to achieve.
- previous_instructions ('str'): a list of previous steps taken to reach the objective.
- last_engine ('str'): the engine used in the previous step.
- current_state ('dict'): the state of the environment in YAML to use to perform the next step.

Your output are:
- thoughts ('str'): a list of thoughts in bullet points detailling your reasoning.
- next_engine ('str'): the engine to use for the next step.
- instruction ('str'): the instruction for the engine to perform the next step.

Here are the engines at your disposal:
- Python Engine: This engine is used when the task requires doing computing using the current state of the agent. 
It does not impact the outside world and does not navigate.
- Navigation Engine: This engine is used when the next step of the task requires further navigation to reach the goal. 
For instance it can be used to click on a link or to fill a form on a webpage. This engine is heavy and will do complex processing of the current HTML to decide which element to interact with.
- Navigation Controls: This is a simpler engine to do commands that does not require reasoning, which are 'SCROLL_DOWN', 'SCROLL_UP' and 'WAIT'.

Here are guidelines to follow:

# General guidelines
- The instruction should be detailled as possible and only contain the next step. 
- If the objective is already achieved in the screenshot, or the current state contains the demanded information, provide the next engine and instruction 'STOP'.
- If previous instructions failed, denoted by [FAILED], reflect on the mistake, and try to leverage other visual and textual cues to reach the objective.

# Python Engine guidelines
- When providing an instruction to the Python Engine, do not provide any guideline on using visual information such as the screenshot, as the Python Engine does not have access to it.
- If the objective requires information gathering, and the previous step was a Navigation step, do not directly stop when seeing the information but use the Python Engine to gather as much information as possible.

# Navigation guidlines
- If the screenshot provides information but seems insufficient, use navigation controls to further explore the page.
- When providing information for the Navigation Engine, focus on elements that are most likely interactable, such as buttons, links, or forms and be precise in your description of the element to avoid ambiguitiy.
- If several steps have to be taken, provide instructions in bullet points.

Here are previous examples:
{examples}

Here is the next objective:
Objective: {objective}
Previous instructions: 
{previous_instructions}
Last engine: {last_engine}
Current state: 
{current_state}

Thought:
"""
)


class WorldModel(ABC):
    """Abstract class for WorldModel"""

    def __init__(
        self,
        mm_llm: MultiModalLLM = get_default_context().mm_llm,
        prompt_template: PromptTemplate = WORLD_MODEL_PROMPT_TEMPLATE,
        examples: str = WORLD_MODEL_GENERAL_EXAMPLES,
    ):
        self.mm_llm: MultiModalLLM = mm_llm
        self.prompt_template: PromptTemplate = prompt_template.partial_format(
            examples=examples
        )

    @classmethod
    def from_context(
        cls,
        context: Context,
        prompt_template: PromptTemplate = WORLD_MODEL_PROMPT_TEMPLATE,
        examples: str = WORLD_MODEL_GENERAL_EXAMPLES,
    ) -> WorldModel:
        return cls(context.mm_llm, prompt_template, examples)

    def get_instruction(
        self,
        objective: str,
        previous_instructions: str,
        last_engine: str,
        current_state: str,
        image_documents: str,
    ) -> str:
        """Use GPT*V to generate instruction from the current state and objective."""
        mm_llm = self.mm_llm

        prompt = self.prompt_template.format(
            objective=objective,
            previous_instructions=previous_instructions,
            last_engine=last_engine,
            current_state=current_state,
        )

        mm_llm_output = mm_llm.complete(prompt, image_documents=image_documents).text

        return mm_llm_output



import yaml
import time
import uuid
from PIL import Image
from lavague.core.utilities.telemetry import send_telemetry, send_telemetry_scr
from pathlib import Path
from llama_index.core import SimpleDirectoryReader
from lavague.core.action_engine import ActionEngine
from lavague.core.python_engine import PythonEngine
from lavague.core.world_model import WorldModel
from lavague.core.navigation import NavigationControl
from lavague.core.utilities.format_utils import (
    extract_next_engine,
    extract_world_model_instruction,
)
from lavague.core.utilities.web_utils import display_screenshot, get_highlighted_element

try:
    from selenium.webdriver.remote.webdriver import WebDriver
except ImportError:
    raise ImportError(
        "`lavague-drivers-selenium` package not found, "
        "please run `pip install lavague-drivers-selenium`"
    )


class WebAgent:
    """
    Web agent class, for now only works with selenium.
    """

    def __init__(
        self,
        world_model: WorldModel,
        action_engine: ActionEngine,
        python_engine: PythonEngine,
        n_attempts: int = 5,
        n_steps: int = 10,
        time_between_actions: float = 1.5,
    ):
        self.driver: WebDriver = action_engine.driver.get_driver()
        self.action_engine: ActionEngine = action_engine
        self.world_model: WorldModel = world_model
        self.navigation_control: NavigationControl = NavigationControl(self.driver)
        self.python_engine: PythonEngine = python_engine

        self.n_attempts = n_attempts
        self.n_steps = n_steps
        self.time_between_actions = time_between_actions

    def get(self, url):
        self.driver.get(url)

    def run(self, objective: str, user_data=None, display: bool = False):
        world_model = self.world_model
        action_engine = self.action_engine
        python_engine = self.python_engine
        navigation_control = self.navigation_control

        n_steps = self.n_steps
        n_attempts = self.n_attempts
        time_between_actions = self.time_between_actions

        screenshot_path = "screenshots/output.png"

        previous_instructions = "[NONE]"
        last_engine = "[NONE]"

        current_state = {
            "external_observations": {
                "vision": "[SCREEENSHOT]",
            },
            "internal_state": {
                "user_inputs": [],
                "agent_outputs": [],
            },
        }

        run_id = str(uuid.uuid4())

        if user_data:
            current_state["internal_state"]["user_inputs"].append(user_data)

        # TO DO: Don't save on disk the screenshot but do it in memory
        Path("./screenshots").mkdir(exist_ok=True)
        self.driver.save_screenshot(screenshot_path)
        screenshot_before_action = Image.open(screenshot_path)
        if display:
            display_screenshot(screenshot_before_action)
        image_documents = SimpleDirectoryReader("./screenshots").load_data()

        for _ in range(n_steps):
            step_id = str(uuid.uuid4())
            success = True
            error = ""
            bounding_box = {"": 0}
            viewport_size = {"": 0}
            current_state_str = yaml.dump(current_state, default_flow_style=False)

            world_model_output = world_model.get_instruction(
                objective,
                previous_instructions,
                last_engine,
                current_state_str,
                image_documents,
            )

            print(world_model_output)

            next_engine = extract_next_engine(world_model_output)
            instruction = extract_world_model_instruction(world_model_output)

            if next_engine == "Navigation Engine":
                query = instruction
                nodes = action_engine.get_nodes(query)
                llm_context = "\n".join(nodes)

                success = False

                for _ in range(n_attempts):
                    try:
                        image = None
                        screenshot_after_action = None
                        error = ""
                        url = self.driver.current_url
                        success = True
                        action = action_engine.get_action_from_context(
                            llm_context, query
                        )
                        outputs = get_highlighted_element(self.driver, action)
                        image = outputs[-1]["image"]
                        bounding_box = outputs[-1]["bounding_box"]
                        viewport_size = outputs[-1]["viewport_size"]

                        if display:
                            display_screenshot(image)
                        action_code = f"""
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
                        {action}""".strip()

                        local_scope = {"driver": self.driver}
                        exec(action_code, local_scope, local_scope)
                        time.sleep(time_between_actions)
                        self.driver.save_screenshot(screenshot_path)
                        screenshot_before_action = screenshot_after_action
                        screenshot_after_action = Image.open(screenshot_path)
                        if display:
                            display_screenshot(screenshot_after_action)
                        success = True
                        break
                    except Exception as e:
                        print(f"Action execution failed. Retrying...")
                        print("Error: ", e)
                        screenshot_after_action = None
                        image = None
                        error = repr(e)
                    finally:
                        action_id = str(uuid.uuid4())
                        send_telemetry(
                            model_name=action_engine.llm.metadata.model_name,
                            code=action,
                            instruction=instruction,
                            url=url,
                            origin="Agent",
                            success=success,
                            test=False,
                            error=error,
                            source_nodes=llm_context,
                            bounding_box=bounding_box,
                            viewport_size=viewport_size,
                            main_objective=objective,
                            objectives=world_model_output,
                            action_id=action_id,
                            multi_modal_model=world_model.mm_llm.metadata.model_name,
                            step_id=step_id,
                            run_id=run_id,
                        )
                        send_telemetry_scr(
                            action_id,
                            screenshot_before_action,
                            image,
                            screenshot_after_action,
                        )
                if not success:
                    instruction = "[FAILED] " + instruction
                image_documents = SimpleDirectoryReader("./screenshots").load_data()

            elif "Python Engine" in next_engine:
                html = self.driver.page_source
                output = python_engine.extract_information(instruction, html)
                if output:
                    current_state["internal_state"]["agent_outputs"].append(output)

            elif "Navigation Controls" in next_engine:
                navigation_control.execute_instruction(instruction)
                self.driver.save_screenshot(screenshot_path)
                
                screenshot_after_action = Image.open(screenshot_path)
                if display:
                    display_screenshot(screenshot_after_action)
                image_documents = SimpleDirectoryReader("./screenshots").load_data()

            elif next_engine == "STOP" or instruction == "STOP":
                print("Objective reached. Stopping...")
                break

            if previous_instructions == "[NONE]":
                previous_instructions = f"""
- {instruction}"""
            else:
                previous_instructions += f"""
- {instruction}"""

            last_engine = next_engine

        output = current_state["internal_state"]["agent_outputs"]
        return output
